import re
import inspect
from typing import Callable
from .token import Token
from .types import TokenTypes


class LambdaTokenizer:
    def __init__(self, expression: str | Callable):
        if callable(expression):
            expression = self.get_expression(expression)
        self.expression = expression
        
        # Create the regular expression pattern to match various tokens
        self.pattern = ""
        self.pattern += r'(\bnot\s+in\b)|' # not in
        self.pattern += r'(\b\w+\.\w+\b|\b\w+\b(?:\(\))?)|' # variable or function
        self.pattern += r'(["\'].*?["\'])|' # string
        self.pattern += r'([<>!=]=?|==|in|and|or)|' # comparison operators
        self.pattern += r'(\d+)|' # number
        self.pattern += r'(\[\])|' # empty list
        self.pattern += r'(\[.*?\])|' # list
        self.pattern += r'(\(.*?\))|' # parentheses
        self.pattern += r'(\s)' # whitespace
        self.pattern = self.pattern.replace("\n", "").replace(" ", "")
       

    @staticmethod
    def get_expression(fnx: Callable) -> str:
        """Extract the lambda expression from the source code"""

        assert callable(fnx) and fnx.__name__ == "<lambda>", "The input must be a lambda expression"
        assert len(inspect.signature(fnx).parameters) == 0, "The lambda expression must not have any arguments"

        # Get the source code of the lambda expression
        source = inspect.getsource(fnx)

        # Extract the lambda expression from the source code
        source = source[source.find("lambda") + len("lambda"):]
        source = source[source.index(":") + 1:]
        source = source.strip()

        # Remove any leading or trailing parentheses
        parentheses = 0
        cleaned_source = ""
        for k in source:
            if k == "(": 
                parentheses += 1

            if k == ")":
                parentheses -= 1

            if k == ")" and parentheses < 0:
                break

            cleaned_source += k
        source = cleaned_source.strip()

        return source

    def get_tokens(self, expression: str = None) -> list[str]:
        """Extract the tokens from the expression"""
        expression = expression or self.expression

        # Regular expression pattern to match various tokens
        tokens = re.findall(self.pattern, expression)
    
        # Check for nested parentheses and flatten the results
        flat_tokens: list[str] = [token for group in tokens for token in group if token.strip()]

        final_tokens = []
        for token in flat_tokens:
            if token.startswith('(') and token.endswith(')'):
                final_tokens += ["(", *self.get_tokens(token[1:-1]) ,")"]
            else:
                final_tokens += [token]
    
        return final_tokens
    
    def tokenize(self):
        """Tokenize the expression"""
        chunks = self.get_tokens()

        tokens = []
        for token in chunks:
            # Check for chain operators
            if token in Token.COMPARISON:
                tokens.append(Token.from_comparison(token))
                continue

            # Check for chain operators
            if token in Token.CHAIN:
                tokens.append(Token.from_chain(token))
                continue
                
            # Check for parentheses
            if token in Token.PARENTHESES:
                tokens.append(Token.from_parentheses(token))
                continue

            # If it's a digit, create a token from it
            if token.isdigit():
                tokens.append(Token(token, TokenTypes.CONSTANT))
                continue
            
            # If it's a string, create a token from it
            if token.startswith('"') or token.startswith("'"):
                tokens.append(Token(token[1:-1], TokenTypes.CONSTANT))
                continue
        

            # Try to load the table class
            if "." in token:  
                elements = token.split(".")  
                try:
                    element = self.run_code_in_frame(elements[0])
                    # assert issubclass(element, Table), f"Class {elements[0]} is not a subclass of Table"
                    tokens.append(Token(elements[1], element))
                    continue
                except AssertionError or KeyError:
                    pass
                
            # If it's not a class, try to get the value from the module where was defined
            code = self.run_code_in_frame(token)
            tokens.append(Token(str(code), TokenTypes.CONSTANT))

        return tokens
    
    def run_code_in_frame(self, token: str, max_depth: int = 10):
        """Get the content of a variable"""
        frame = inspect.currentframe().f_back
        string = token.split(".")[0] if "." in token else token

        # Get the module where the lambda expression is defined
        for _ in range(max_depth):
            if string in frame.f_locals:
                break
        
            if string in frame.f_globals:
                break

            frame = frame.f_back
        
        assert string in frame.f_locals or string in frame.f_globals, f"Variable {string} not found in the module"
       
        # Use the module to execute the code
        frame_globals = frame.f_globals
        frame_locals = frame.f_locals

        # Prepare the context to execute the code
        full_code = f'__return_value__ = {token}'

        # Execute the code in the frame context
        exec(full_code, frame_globals, frame_locals)

        # clean up the frame
        return_value = frame_locals['__return_value__']
        del frame_locals['__return_value__']

        return return_value
        





import inspect
from typing import TYPE_CHECKING
from ..tokenizer.types import TokenTypes

if TYPE_CHECKING:    
    from ..table import Table
    from ..tokenizer.token import Token

import sqlite3
from .__template__ import DbProvider

TYPES_SQLITE = {
    int: 'INTEGER',
    str: 'TEXT',
    float: 'REAL',
    bool: 'INTEGER',
}


class Sqlite(DbProvider):
    def __init__(self, path: str):
        self.path = path
        self.name = path.split("/")[-1]
        self.conn = sqlite3.connect(path)
        self.cursor = None

        # Fetch the existing tables
        cursor = self.conn.execute("SELECT name FROM sqlite_master WHERE type='table'")
        self.__tables__ = [table[0] for table in cursor.fetchall()]
        
    def _get_schema(self, table: "Table") -> dict:
        """Return the schema of the table"""
        if not table.__exist__:
            return {}
        
        cursor = self.conn.execute(f"PRAGMA table_info({table.__table__})")
        schema = {row[1]: row[2] for row in cursor.fetchall()}
        return schema

    def create_table(self, table: "Table"):
        """Create the table if it does not exist"""
        # Check if the table already exists
        if table.__exist__:
            return

        types = [f"{key} {self._convert_type(value)}" for key, value in table.__annotations__.items()]
        schema = f"CREATE TABLE {table.__table__} ({', '.join(types)})"
        self.conn.execute(schema)
        self.conn.commit()
        table.__exist__ = True
        table.__schema__ = self._get_schema(table)
        self.__tables__.append(table.__table__)

    def drop_table(self, table: "Table"):
        """Drop the table from the database"""
        if not table.__exist__:
            return
        
        self.conn.execute(f"DROP TABLE {table.__table__}")
        self.conn.commit()
        table.__exist__ = False
    
    def update_table(self, table: "Table"):
        """Update the table schema"""
        if not table.__exist__:
            return

        # Add the missing columns
        missing_columns = self._get_missing_columns(table)
        for column in missing_columns:
            column_type = self._convert_type(table.__annotations__[column])
            self.conn.execute(f"ALTER TABLE {table.__table__} ADD COLUMN {column} {column_type}")
            self.conn.commit()

        # Remove the extra columns
        extra_columns = self._get_extra_columns(table)
        for column in extra_columns:
            self.conn.execute(f"ALTER TABLE {table.__table__} DROP COLUMN {column}")
            self.conn.commit()

        # Update the schema
        table.__schema__ = self._get_schema(table)

    def commit(self):
        """Commit the changes to the database"""
        self.conn.commit()
        
    def close(self):
        """Close the connection to the database"""
        self.conn.close()
    ##########################################
    # Database operations
    ##########################################
    def _insert(self, obj: dict, table: str):
        """Insert a row into the table"""
        columns = ', '.join(obj.keys())
        values = ', '.join([f"'{value}'" if isinstance(value, str) else str(value) for value in obj.values()])
        self.conn.execute(f"INSERT INTO {table} ({columns}) VALUES ({values})")
        # self.conn.commit()

    def execute(self, query: str):
        """Select the table from the database"""
        if self.cursor:
            self.cursor.execute(query)
        cursor = self.conn.execute(query)
        return cursor.fetchall()

    ##########################################
    # Utility functions
    ##########################################
    def _convert_type(self, type):
        """Convert the type to the sqlite type""" 
        assert type in TYPES_SQLITE, f"Type {type} not supported"
        return TYPES_SQLITE[type]
    
    def _get_missing_columns(self, table: "Table") -> list:
        """Return the missing columns"""
        return [key for key in table.__annotations__ if key not in table.__schema__]
    
    def _get_extra_columns(self, table: "Table") -> list:
        """Return the extra columns"""
        return [key for key in table.__schema__ if key not in table.__annotations__]
    
    def _get_changed_columns(self, table: "Table") -> list:
        """Return the changed columns"""
        return [key for key in table.__annotations__ if key in table.__schema__ and table.__schema__[key] != self._convert_type(table.__annotations__[key])]
    
    def _tokens_to_sql(self, tokens: list["Token"], table: "Table") -> str:
        """Convert the token to a sql string"""
        # Get the table order of elements
        columns = list(table.__annotations__.keys())
        query = "SELECT"
        for column in columns:
            query += " %s," % column
        query = query[:-1] + " FROM %s" % table.__table__

        if not tokens:
            return query
        else:
            query += " WHERE"

        for token in tokens:
            query += " "
            print(token)
            if inspect.isclass(token.token_type):
                query += "%s" % token.value
                continue
            
            match token.token_type:
                case TokenTypes.AND:
                    query += "AND"
                case TokenTypes.OR:
                    query += "OR"
                case TokenTypes.OPEN_PARENTHESIS:
                    query += "("
                case TokenTypes.CLOSE_PARENTHESIS:
                    query += ")"
                case TokenTypes.LESS_THAN:
                    query += "<"
                case TokenTypes.GREATER_THAN:
                    query += ">"
                case TokenTypes.LESS_THAN_EQUAL:
                    query += "<="
                case TokenTypes.GREATER_THAN_EQUAL:
                    query += ">="
                case TokenTypes.EQUAL:
                    query += "="
                case TokenTypes.NOT_EQUAL:
                    query += "!="
                case TokenTypes.IN:
                    query += "IN"
                case TokenTypes.NOT_IN:
                    query += "NOT IN"
                case TokenTypes.CONSTANT:
                    query += "'%s'" % token.value
                
                case Token:
                    query += f" {token.value}"

        return query

    ##########################################
    # Magic methods
    ##########################################
    def __del__(self):
        self.close()

    def __enter__(self):
        self.cursor = self.conn.cursor()
        return self.cursor
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type or exc_val or exc_tb:
            # If there was an exception, rollback any changes
            self.conn.rollback()
        else:
            # Otherwise, commit the changes
            self.conn.commit()

        if self.cursor:
            self.cursor.close()


class TokenTypes:
    """A class to hold the various token types"""
    # Chain operators
    AND = "AND" # and
    OR = "OR" # or

    # Comparison operators
    EQUAL = "EQUAL" # ==
    NOT_EQUAL = "NOT_EQUAL" # !=
    LESS_THAN = "LESS_THAN" # <
    GREATER_THAN = "GREATER_THAN" # >
    LESS_THAN_EQUAL = "LESS_THAN_EQUAL" # <=
    GREATER_THAN_EQUAL = "GREATER_THAN_EQUAL" # >=
    IN = "IN" # in
    NOT_IN = "NOT_IN" # not in

    # Variables
    VARIABLE = "VARIABLE" # variable or function
    CONSTANT = "CONSTANT" # number or string
    TABLE = "TABLE" # table

    # Parentheses
    OPEN_PARENTHESIS = "OPEN_PARENTHESIS" # (
    CLOSE_PARENTHESIS = "CLOSE_PARENTHESIS" # )


from abc import ABC, abstractmethod
from dataclasses import dataclass

from typing import dataclass_transform, TYPE_CHECKING
if TYPE_CHECKING:
    from ..table import Table

class DbProvider(ABC):
    """
        Abstract class for the database provider
        Each provider extend this class and implement the methods that are provider specific
    """
    __tables__ = []
    __classes__ = []

    @dataclass_transform()
    def useTable(self, table_name: str):
        def decorator(cls: "Table"):
            # Apply the dataclass decorator
            cls = dataclass(cls)

            # Check that all the keys have a type and that the value is of the right type
            for key in cls.__dict__:
                if key.startswith("__"):
                    continue
                assert key in cls.__annotations__, f"Key '{key}' does not have a type in the class {cls.__name__}"
                assert isinstance(cls.__dict__[key], cls.__annotations__[key]), f"Value {cls.__dict__[key]} is not of type {cls.__annotations__[key]}"

            # Assign the table name to the class
            cls.__table__ = table_name

            # Assign the current provider to the class
            cls.__provider__ = self

            # Check if the table exists
            if table_name in self.__tables__:
                cls.__exist__ = True
                cls.__schema__ = self._get_schema(cls)

            # Check that columns have the right type
            for key in cls.__annotations__:
                if key in cls.__schema__:
                    msg = f"Column {key} has type {cls.__annotations__[key]} in the class but has type {cls.__schema__[key]} in the database"
                    assert self._convert_type(cls.__annotations__[key]) == cls.__schema__[key],msg

            self.__classes__.append(cls)
            return cls
        return decorator

    @abstractmethod
    def _get_schema(self, table) -> dict:
        """Return the schema of the table"""

    @abstractmethod
    def create_table(self, table):
        """Create a table in the database"""

    def create_tables(self):
        """Create the table in the database"""
        for table in self.__classes__:
            self.create_table(table)

    @abstractmethod
    def drop_table(self, table):
        """Drop the table from the database"""

    def drop_tables(self):
        """Drop the table from the database"""
        for table in self.__classes__:
            self.drop_table(table)

    @abstractmethod
    def update_table(self, table):
        """Update the table schema in the database"""

    def update_tables(self):
        """Update the table schema in the database"""
        for table in self.__classes__:
            self.update_table(table)

    @abstractmethod
    def commit(self):
        """Commit the changes to the database"""

    @abstractmethod
    def close(self):
        """Close the connection to the database"""

    ##########################################
    # Database operations
    ##########################################
    @abstractmethod
    def _insert(self, obj, table):
        """Insert a row into the table"""

    @abstractmethod
    def execute(self, query):
        """Select the table from the database"""

    ##########################################
    # Utility functions
    ##########################################
    @abstractmethod
    def _convert_type(self, type) -> str:
        """Convert the type to the database type"""

    @abstractmethod
    def _get_missing_columns(self, table) -> list:
        """Return the missing columns"""

    @abstractmethod
    def _get_extra_columns(self, table) -> list:
        """Return the extra columns"""

    @abstractmethod
    def _get_changed_columns(self, table) -> list:
        """Return the changed columns"""

    @abstractmethod
    def _tokens_to_sql(self, tokens, table) -> str:
        """Convert the token to a sql string"""

    ##########################################
    # Magic methods
    ##########################################
    @abstractmethod
    def __del__(self):
        """Close the connection to the database"""

    @abstractmethod
    def __enter__(self):
        """Enter the context"""

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context"""


import inspect
from .types import TokenTypes


class Token:
    CHAIN = ["and", "or"]
    COMPARISON = ["<", ">", "<=", ">=", "==", "!=", "in", "not in"]
    PARENTHESES = ["(", ")"]

    def __init__(self, value: str, token_type: str):
        # A token can be a Table child class or a token type
        self.value = value
        self.token_type = token_type

    @classmethod
    def from_chain(cls, value: str) -> "Token":
        """Create a token from a chain operator"""
        match value:
            case "and":
                return cls(value, TokenTypes.AND)
            case "or":
                return cls(value, TokenTypes.OR)
            case _:
                raise ValueError(f"Invalid chain operator: {value}")

    @classmethod
    def from_comparison(cls, value: str) -> "Token":
        """Create a token from a comparison operator"""
        match value:
            case "<":
                return cls(value, TokenTypes.LESS_THAN)
            case ">":
                return cls(value, TokenTypes.GREATER_THAN)
            case "<=":
                return cls(value, TokenTypes.LESS_THAN_EQUAL)
            case ">=":
                return cls(value, TokenTypes.GREATER_THAN_EQUAL)
            case "==":
                return cls(value, TokenTypes.EQUAL)
            case "!=":
                return cls(value, TokenTypes.NOT_EQUAL)
            case "in":
                return cls(value, TokenTypes.IN)
            case "not in":
                return cls(value, TokenTypes.NOT_IN)
            case _:
                raise ValueError(f"Invalid comparison operator: {value}")
        
    @classmethod
    def from_parentheses(cls, value: str) -> "Token":
        """Create a token from a parentheses"""
        match value:
            case "(":
                return cls(value, TokenTypes.OPEN_PARENTHESIS)
            case ")":
                return cls(value, TokenTypes.CLOSE_PARENTHESIS)
            case _:
                raise ValueError(f"Invalid parentheses: {value}")

    ##################################################
    # Magic methods
    ##################################################
    def __repr__(self):
        # Check if the token is a Table child
        if inspect.isclass(self.token_type):
            return f"{self.token_type.__name__}('{self.value}')"
        
        return f"{self.token_type.capitalize()}('{self.value}')"
  
        
    def __eq__(self, other: "Token"):
        return self.value == other.value and self.token_type == other.token_type


from typing import TYPE_CHECKING
from .tokenizer.tokenizer import LambdaTokenizer

if TYPE_CHECKING:
    from .providers.__template__ import DbProvider


class Table:
    __exist__: bool = False # If the table exists
    __table__: str = None # The table name
    __schema__: dict = {} # Key value pairs of the schema
    __provider__: "DbProvider" = None # The provider

    def __init__(self, **kwargs):
        # Check if the table name is set
        assert self.__table__ is not None, "Table name not set, create a connector and use the useTable decorator"

        # Assign the values to the object
        for key, value in kwargs.items():
            assert key in self.__annotations__, f"Key {key} not in table {self.__table__}"
            assert isinstance(value, self.__annotations__[key]), f"Value {value} is not of type {self.__annotations__[key]}"
            setattr(self, key, value)

        # Check if all the keys are set
        for key in self.__annotations__:
            assert hasattr(self, key), f"Key {key} not set for table {self.__table__}"

    def to_dict(self):
        """Return the object as a dictionary"""
        return {key: getattr(self, key) for key in self.__annotations__}
    
    def save(self):
        """Save the object to the database"""
        assert self.__exist__, f"Table {self.__table__} does not exist, create it first using the create_table method"  
        self.__provider__._insert(self.to_dict(), self.__table__)

    ##########################################
    # CQuery database methods
    ##########################################
    @classmethod
    def where(cls, *args, **kwargs):
        """Return the query"""
        # If no arguments are passed, select all
        if not args and not kwargs:
            query = cls.__provider__._tokens_to_sql([], cls)

        # If the first argument is a lambda function
        if len(args) == 1 and callable(args[0]):
            tokens = LambdaTokenizer(args[0]).tokenize()
            query = cls.__provider__._tokens_to_sql(tokens, cls)

        if not query:
            return []
        
        data = cls.__provider__.execute(query)
        columns = list(cls.__annotations__.keys())
        data = [dict(zip(columns, row)) for row in data]
        return [cls(**row) for row in data]

    ##########################################
    # Magic method
    ##########################################
    def __repr__(self):
        content = ', '.join([f'{key}={getattr(self, key)}' for key in self.__annotations__]) 
        return f"{self.__class__.__name__}({content})"
    
    def __getitem__(self, key):
        """Get the value of the key"""
        assert key in self.__annotations__, f"Key {key} not in table {self.__table__}"
        return getattr(self, key)
    
    def __setitem__(self, key, value):
        """Set the value of the key"""
        assert key in self.__annotations__, f"Key {key} not in table {self.__table__}"
        assert isinstance(value, self.__annotations__[key]), f"Value {value} is not of type {self.__annotations__[key]}"
        setattr(self, key, value)

    def __len__(self):
        """Return the number of columns"""
        return len(self.__annotations__)
    
    def __contains__(self, key):
        """Check if the key is in the table"""
        return key in self.__annotations__
    
    def __eq__(self, other):
        """Check if the two objects are equal"""
        return all([getattr(self, key) == getattr(other, key) for key in self.__annotations__])